/*
 *  gbe - gameboy emulator
 *  Copyright (C) 1999  Chuck Mason, Steven Fuller
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 *
 *  Chuck Mason <chuckjr@sinclair.net>
 *  Steven Fuller <relnev@atdot.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>

#include "getopt.h"

#include "rom.h"
#include "mem.h"
#include "data.h"
#include "vram.h"
#include "cpu.h"
#include "joypad.h"
#include "debug.h"

#define OPTSTRING "dhvc:j:m:s0:1:2:3:f:"

struct option long_options[] = {
	{ "debug",	0,	NULL,	'd' },
	{ "version",	0,	NULL,	'v' },
	{ "cycles",	1,	NULL,	'c' },
	{ "joystick", 	1, 	NULL,	'j' },
	{ "multiply",	1,	NULL, 	'm' },
	{ "svga",	0,	NULL,	's' },
	{ "color0",	1,	NULL,	'0' },
	{ "color1",	1,	NULL,	'1' },
	{ "color2",	1,	NULL,	'2' },
	{ "color3",	1,	NULL,	'3' },
	{ "force",	1,	NULL,	'f' },
	0
};

gameboy_proc_t *current_processor = NULL;
unsigned long cycles_per_hblank   = 451;
unsigned char use_joystick	  = 5;
unsigned char gameboy_screen_mul  = 1;
force_type force_system           = NONE;

void sigint(int sig);

/* 
 * TODO: Change -s to -s n where n is graphic mode
 * (and list out available graphic modes in help)
 */
 
void help(char *name)
{
	printf("\nusage: %s [OPTIONS] ... FILENAME\n\n", name);
	printf("\t-d\tenter debugger immediately\n"
	       "\t-v\tdisplay version and quit\n"
	       "\t-c n\tset cycles per hblank to n\n"
#ifdef BUILD_JLIB
	       "\t-j n\tuse joystick device /dev/jsN\n"
#elif defined (WIN32)
	       "\t-j\tuse joystick\n"
#endif
	       "\t-m n\tset screen multiply to n\n"
#ifdef BUILD_SVGA
	       "\t-s\tuse SVGAlib instead of X11 drivers\n"
#endif
	       "\t-[0-3] x\tset color [0-3] to x (where X can be prefixed with 0x)\n"
	       "\t-f [color|regular] force emulation of color or regular gameboy\n"
	);
}

int main(int argc, char *argv[])
{
	char optch;
	char enter_debugger = 0;
	
	printf("gbe v0.0.22: gameboy emulator by\n");
	printf("\tChuck Mason <chuckjr@sinclair.net>\n");
	printf("\tSteven Fuller <relnev@atdot.org>\n");

	if(argc < 2) {
	        help(argv[0]);
		return 0;
	}

	while((optch = getopt_long(argc, argv, OPTSTRING, long_options, NULL)) != EOF) {
		switch(optch) {
			case 'd':
				enter_debugger = 1;
				break;
			case 'h':
				help(argv[0]);
			case 'v':
				exit(EXIT_SUCCESS);
				break;
			case 'c':
				cycles_per_hblank = strtoul(optarg, NULL, 10);
				printf("Cycles per HBlank set to %d\n", cycles_per_hblank);
				break;
			case 'j':
				use_joystick = strtoul(optarg, NULL, 10);
				printf("Using joystick device #%d\n", use_joystick);
				break;
			case 'm':
				gameboy_screen_mul = strtoul(optarg, NULL, 10);
				printf("Screen multiply set to %d\n", gameboy_screen_mul);
				break;

#ifdef BUILD_SVGA
			/* TODO: (see TODO note above about graphics modes) */
			case 's':
				printf("Using SVGAlib...\n");
				graphic_sel = 1;
				if(gameboy_screen_mul > 1) {
					printf("Screen multiply only available in X\n");
					gameboy_screen_mul = 1;
				}
				break;
#endif
			case '0':
				vram_set_color(0, strtoul(optarg, NULL, 0));
				break;
			case '1':
				vram_set_color(1, strtoul(optarg, NULL, 0));
				break;
			case '2':
				vram_set_color(2, strtoul(optarg, NULL, 0));
				break;
			case '3':
				vram_set_color(3, strtoul(optarg, NULL, 0));
				break;
			case 'f':
				if(strcmp(optarg, "color") == 0)
					force_system = COLOR;
				else if(strcmp(optarg, "regular") == 0)
					force_system = REGULAR;
				else {
					printf("unknown system %s\n", optarg);
					return 0;
				}
				break;	
			case ':':
				exit(EXIT_FAILURE);
				break;
			case '?':
				exit(EXIT_FAILURE);
				break;
		}
	}

        optch = argc - optind;
        switch(optch) {
                case 0:
                        fprintf(stderr, "gbe: no gameboy image specified.\n");
                        exit(0);
                case 1:
                        break;
                default:
                        fprintf(stderr, "gbe: too many files specified at command line\n");
                        exit(0);
        }

	if(initialize_memory() == 0) {
		free_memory();	/* Free everything that was 'actually' allocated.. */
		
		printf("Unable to allocate memory\n");
		return 0;
	}
		
	if(load_rom(argv[optind])) {
		gbe_init_gfx();
		joypad_init();
		
		signal(SIGINT, sigint);
		
		if(current_processor = gameboy_cpu_hardreset()) {
			if(enter_debugger)
				gbz80_debugger(current_processor);
			else
				gameboy_cpu_run(current_processor);	
		}
		gbe_close_gfx();
	} else {
		fprintf(stderr, "Unable to load '%s'\n", argv[optind]);
	}

	free_rom();
	free_memory();

	if(current_processor) {
		free(current_processor);
	}

	return 1;
}


void sigint(int sig)
{
#if 0
	FILE *fp;
	int i;

	fp = fopen("vram.dump", "w");
	for(i = 0; i < 0x2000; i++)
		fputc(memory_read_byte(0x8000+i), fp);
	fclose(fp);
	current_processor->running = 0;
#endif
	signal(SIGINT, sigint);	
	gbz80_debugger(current_processor);
}
