/* this file is too, released under GPL. 
 * this file is maintained by Jeff Miller <miller@zipcon.net>
 */

#ifdef BUILD_WIN32

#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include <stdio.h>
#include <malloc.h>
#include "joypad.h"

#define GFX_BITDEPTH 16
// This should always be ( GFX_BITDEPTH / 8 )
#define GFX_BYTEDEPTH 2

static unsigned long gbe_window_width = 0;
static unsigned long gbe_window_height = 0;

// This will be set by CreateDIBSection
static unsigned char* gfx_buffer = NULL;

static const char* CLASS_NAME = "GBE_CLASS";
HWND g_hwnd = NULL;
HDC g_hMainDC = NULL;
BITMAPINFOHEADER* g_pDib = NULL;
HBITMAP g_hDibSect = NULL;

static LRESULT CALLBACK WindowProc(HWND, UINT, WPARAM, LPARAM);

static UINT uFrameCount;
static unsigned __int64 beginTime;
static unsigned __int64 ticksPerSecond;

int gbe_init_win32(char *display)
{

    // Get the initial tick count
    QueryPerformanceCounter((LARGE_INTEGER*)&beginTime);
	uFrameCount = 0;
    // Find out how many ticks occur per second
    QueryPerformanceFrequency((LARGE_INTEGER*)&ticksPerSecond);

    return GFX_BYTEDEPTH;
}

int gbe_close_win32()
{
    unsigned __int64 endTime;
    DWORD secs;
    DWORD fps;
    TCHAR tszBuf[1024];

    if ( g_pDib )
    {
        free( g_pDib );
    }

    if ( g_hDibSect )
    {
        DeleteObject( g_hDibSect );
    }

    // Get the final tick count
    QueryPerformanceCounter((LARGE_INTEGER*)&endTime);
    // Get number of ticks
    secs = (DWORD)((endTime - beginTime) / ticksPerSecond);
    fps = (secs == 0) ? 0 : (uFrameCount / secs);
    wsprintf(tszBuf, "Frames drawn: %ld\nFPS: %ld", uFrameCount, fps);
    MessageBox(NULL, tszBuf, "Statistics", MB_OK);

    return 1;
}



int gbe_create_window_win32(char *title, int width, int height)
{
    HINSTANCE hInstance = GetModuleHandle( NULL );
    WNDCLASS wc;

    memset(&wc, 0, sizeof(wc));
    wc.hCursor = LoadCursor( NULL, IDC_ARROW );
    wc.hIcon = LoadIcon( NULL, IDI_ASTERISK);
    wc.hbrBackground = (HBRUSH)( COLOR_WINDOW + 1 );
    wc.hInstance = hInstance;
    wc.lpfnWndProc = WindowProc;
    wc.lpszClassName = CLASS_NAME;
    if (RegisterClass(&wc) == 0)
    {
        return 0;
    }
    
    g_hwnd = CreateWindow(CLASS_NAME, "gbe",
        WS_POPUPWINDOW | WS_CAPTION | WS_VISIBLE,
        CW_USEDEFAULT, CW_USEDEFAULT,
        width + (2 * GetSystemMetrics(SM_CXFIXEDFRAME)), height
            + (2 * GetSystemMetrics(SM_CYFIXEDFRAME)
            + GetSystemMetrics(SM_CYMENU)),
        HWND_DESKTOP,
        NULL,
        hInstance,
        NULL);
    if (g_hwnd == NULL)
    {
        UnregisterClass(CLASS_NAME, hInstance);
        return 0;
    }

    gbe_window_width = width;
    gbe_window_height = height;

    g_hMainDC = GetDC(g_hwnd);

    g_pDib = (BITMAPINFOHEADER*)malloc(
        sizeof( BITMAPINFOHEADER ) + 256 * sizeof( RGBQUAD ) );
    if ( g_pDib == NULL )
    {
        return 0;
    }

    memset( g_pDib, 0, sizeof( BITMAPINFOHEADER ) );
    g_pDib->biSize = sizeof( BITMAPINFOHEADER );
    g_pDib->biWidth = width;
    g_pDib->biHeight = -height; // top down
    g_pDib->biPlanes = 1;
    g_pDib->biBitCount = GFX_BITDEPTH; // 16 bit color
    g_pDib->biCompression = BI_RGB;

    g_hDibSect = CreateDIBSection( g_hMainDC, (BITMAPINFO*)g_pDib,
        DIB_RGB_COLORS, &gfx_buffer, NULL, 0);
    if ( g_hDibSect == NULL )
    {
        return 0;
    }

    ShowWindow(g_hwnd, SW_SHOWDEFAULT);
    
    return 1;
}

static void keyboard_handler(WPARAM code, BOOL press)
{
    int button = 0;

    switch ( code )
    {
    case VK_RETURN:
        button = GBE_START;
        break;

    case VK_TAB:
        button = GBE_SELECT;
        break;

    case VK_CONTROL:
        button = GBE_B;
        break;

    case VK_SPACE:
        button = GBE_A;
        break;

    case VK_LEFT:
        button = GBE_LEFT;
        break;

    case VK_RIGHT:
        button = GBE_RIGHT;
        break;

    case VK_UP:
        button = GBE_UP;
        break;

    case VK_DOWN:
        button = GBE_DOWN;
        break;

    }

    if ( button )
    {
        if ( press )
        {
            joypad_press( button );
        }
        else
        {
            joypad_release( button );
        }
    }
}


int gbe_blit_win32()
{
    SetDIBitsToDevice( g_hMainDC, 
        0, 0, 
        gbe_window_width, gbe_window_height, 
        0, 0, 0, gbe_window_height,
        gfx_buffer, (BITMAPINFO*)g_pDib, DIB_RGB_COLORS );

    ++uFrameCount;

    return 1;
}

int gbe_poll_events_win32()
{
    static MSG msg;

    if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
    {
        switch (msg.message)
        {
        case WM_ERASEBKGND:
            return 1;

        case WM_QUIT:
            joypad_press(GBE_O_ESCAPE);
            joypad_release(GBE_O_ESCAPE);
            return 0;

        case WM_KEYDOWN:
        case WM_KEYUP:
            keyboard_handler(msg.wParam, msg.message == WM_KEYDOWN);
            return 1;
        }
        
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }
    
    return 1;
}

int gbe_plot_win32(int x, int y, int color)
{
#if GFX_BYTEDEPTH == 2

    gfx_buffer[y * gbe_window_width * GFX_BYTEDEPTH + x * GFX_BYTEDEPTH] 
        = (color >> 8) & 0xFF;
    gfx_buffer[y * gbe_window_width * GFX_BYTEDEPTH + x * GFX_BYTEDEPTH + 1] = color & 0xFF;

#else

    gfx_buffer[y * gbe_window_width + x] = color & 0xFF;

#endif

    return 1;
}

// local

static LRESULT CALLBACK WindowProc
(
 HWND hWnd, 
 UINT message, 
 WPARAM wParam, 
 LPARAM lParam
 )
{
    static HINSTANCE   hInstance;
    
    switch( message )
    {
    case WM_CREATE:
        hInstance = ((LPCREATESTRUCT) lParam)->hInstance;
        return 0;
        
    case WM_DESTROY:
        ReleaseDC(hWnd, g_hMainDC);
        PostQuitMessage(0);
        break;

    }
    
    return DefWindowProc(hWnd, message, wParam, lParam);
}

#else

int gbe_init_win32(char *display)
{
    // 0 = mode not supported
	return 0;
}

int gbe_close_win32()
{
	return 1;
}

int gbe_create_window_win32(char *title, int width, int height)
{
		
	return 1;
}


int gbe_poll_events_win32()
{
	return 1;
}

int gbe_blit_win32()
{
	return 1;
}

int gbe_plot_win32(int x, int y, int color)
{
	return 1;
}

#endif
